
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "pos-enum-types.h"
#include "phosh-osk-enums.h"
#include "pos-enums.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "phosh-osk-enums.h" */

GType
phosh_osk_completion_mode_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (PHOSH_OSK_COMPLETION_MODE_NONE), "PHOSH_OSK_COMPLETION_MODE_NONE", "none" },
        { C_FLAGS (PHOSH_OSK_COMPLETION_MODE_MANUAL), "PHOSH_OSK_COMPLETION_MODE_MANUAL", "manual" },
        { C_FLAGS (PHOSH_OSK_COMPLETION_MODE_HINT), "PHOSH_OSK_COMPLETION_MODE_HINT", "hint" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("PhoshOskCompletionModeFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phosh_osk_features_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (PHOSH_OSK_FEATURE_KEY_DRAG), "PHOSH_OSK_FEATURE_KEY_DRAG", "key-drag" },
        { C_FLAGS (PHOSH_OSK_FEATURE_KEY_INDICATOR), "PHOSH_OSK_FEATURE_KEY_INDICATOR", "key-indicator" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("PhoshOskFeatures"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phosh_osk_completion_source_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (PHOSH_OSK_COMPLETION_SOURCE_NONE), "PHOSH_OSK_COMPLETION_SOURCE_NONE", "none" },
        { C_FLAGS (PHOSH_OSK_COMPLETION_SOURCE_EMOJI), "PHOSH_OSK_COMPLETION_SOURCE_EMOJI", "emoji" },
        { C_FLAGS (PHOSH_OSK_COMPLETION_SOURCE_KEYWORD), "PHOSH_OSK_COMPLETION_SOURCE_KEYWORD", "keyword" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("PhoshOskCompletionSourceFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
phosh_osk_scaling_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (PHOSH_OSK_SCALING_AUTO_PORTRAIT), "PHOSH_OSK_SCALING_AUTO_PORTRAIT", "auto-portrait" },
        { C_FLAGS (PHOSH_OSK_SCALING_AUTO_LANDSCAPE), "PHOSH_OSK_SCALING_AUTO_LANDSCAPE", "auto-landscape" },
        { C_FLAGS (PHOSH_OSK_SCALING_BOTTOM_DEAD_ZONE), "PHOSH_OSK_SCALING_BOTTOM_DEAD_ZONE", "bottom-dead-zone" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("PhoshOskScalingFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "pos-enums.h" */

GType
pos_osk_widget_layer_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (POS_OSK_WIDGET_LAYER_NORMAL), "POS_OSK_WIDGET_LAYER_NORMAL", "layer-normal" },
        { C_ENUM (POS_OSK_WIDGET_LAYER_CAPS), "POS_OSK_WIDGET_LAYER_CAPS", "layer-caps" },
        { C_ENUM (POS_OSK_WIDGET_LAYER_SYMBOLS), "POS_OSK_WIDGET_LAYER_SYMBOLS", "layer-symbols" },
        { C_ENUM (POS_OSK_WIDGET_LAYER_SYMBOLS2), "POS_OSK_WIDGET_LAYER_SYMBOLS2", "layer-symbols2" },
        { C_ENUM (POS_OSK_WIDGET_LAST_LAYER), "POS_OSK_WIDGET_LAST_LAYER", "last-layer" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PosOskWidgetLayer"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
pos_osk_widget_mode_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (POS_OSK_WIDGET_MODE_KEYBOARD), "POS_OSK_WIDGET_MODE_KEYBOARD", "keyboard" },
        { C_ENUM (POS_OSK_WIDGET_MODE_CURSOR), "POS_OSK_WIDGET_MODE_CURSOR", "cursor" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PosOskWidgetMode"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
pos_osk_key_use_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (POS_OSK_KEY_USE_KEY), "POS_OSK_KEY_USE_KEY", "key" },
        { C_ENUM (POS_OSK_KEY_USE_DELETE), "POS_OSK_KEY_USE_DELETE", "delete" },
        { C_ENUM (POS_OSK_KEY_USE_TOGGLE), "POS_OSK_KEY_USE_TOGGLE", "toggle" },
        { C_ENUM (POS_OSK_KEY_USE_MENU), "POS_OSK_KEY_USE_MENU", "menu" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PosOskKeyUse"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
pos_input_method_text_change_cause_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (POS_INPUT_METHOD_TEXT_CHANGE_CAUSE_IM), "POS_INPUT_METHOD_TEXT_CHANGE_CAUSE_IM", "im" },
        { C_ENUM (POS_INPUT_METHOD_TEXT_CHANGE_CAUSE_NOT_IM), "POS_INPUT_METHOD_TEXT_CHANGE_CAUSE_NOT_IM", "not-im" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PosInputMethodTextChangeCause"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
pos_input_method_purpose_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_NORMAL), "POS_INPUT_METHOD_PURPOSE_NORMAL", "normal" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_ALPHA), "POS_INPUT_METHOD_PURPOSE_ALPHA", "alpha" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_DIGITS), "POS_INPUT_METHOD_PURPOSE_DIGITS", "digits" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_NUMBER), "POS_INPUT_METHOD_PURPOSE_NUMBER", "number" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_PHONE), "POS_INPUT_METHOD_PURPOSE_PHONE", "phone" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_URL), "POS_INPUT_METHOD_PURPOSE_URL", "url" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_EMAIL), "POS_INPUT_METHOD_PURPOSE_EMAIL", "email" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_NAME), "POS_INPUT_METHOD_PURPOSE_NAME", "name" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_PASSWORD), "POS_INPUT_METHOD_PURPOSE_PASSWORD", "password" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_PIN), "POS_INPUT_METHOD_PURPOSE_PIN", "pin" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_DATE), "POS_INPUT_METHOD_PURPOSE_DATE", "date" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_TIME), "POS_INPUT_METHOD_PURPOSE_TIME", "time" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_DATETIME), "POS_INPUT_METHOD_PURPOSE_DATETIME", "datetime" },
        { C_ENUM (POS_INPUT_METHOD_PURPOSE_TERMINAL), "POS_INPUT_METHOD_PURPOSE_TERMINAL", "terminal" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PosInputMethodPurpose"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
pos_input_method_hint_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (POS_INPUT_METHOD_HINT_NONE), "POS_INPUT_METHOD_HINT_NONE", "none" },
        { C_ENUM (POS_INPUT_METHOD_HINT_COMPLETION), "POS_INPUT_METHOD_HINT_COMPLETION", "completion" },
        { C_ENUM (POS_INPUT_METHOD_HINT_SPELLCHECK), "POS_INPUT_METHOD_HINT_SPELLCHECK", "spellcheck" },
        { C_ENUM (POS_INPUT_METHOD_HINT_AUTO_CAPITALIZATION), "POS_INPUT_METHOD_HINT_AUTO_CAPITALIZATION", "auto-capitalization" },
        { C_ENUM (POS_INPUT_METHOD_HINT_LOWERCASE), "POS_INPUT_METHOD_HINT_LOWERCASE", "lowercase" },
        { C_ENUM (POS_INPUT_METHOD_HINT_UPPERCASE), "POS_INPUT_METHOD_HINT_UPPERCASE", "uppercase" },
        { C_ENUM (POS_INPUT_METHOD_HINT_TITLECASE), "POS_INPUT_METHOD_HINT_TITLECASE", "titlecase" },
        { C_ENUM (POS_INPUT_METHOD_HINT_HIDDEN_TEXT), "POS_INPUT_METHOD_HINT_HIDDEN_TEXT", "hidden-text" },
        { C_ENUM (POS_INPUT_METHOD_HINT_SENSITIVE_DATA), "POS_INPUT_METHOD_HINT_SENSITIVE_DATA", "sensitive-data" },
        { C_ENUM (POS_INPUT_METHOD_HINT_LATIN), "POS_INPUT_METHOD_HINT_LATIN", "latin" },
        { C_ENUM (POS_INPUT_METHOD_HINT_MULTILINE), "POS_INPUT_METHOD_HINT_MULTILINE", "multiline" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PosInputMethodHint"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
pos_backspace_mode_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (POS_BACKSPACE_MODE_CHAR), "POS_BACKSPACE_MODE_CHAR", "char" },
        { C_ENUM (POS_BACKSPACE_MODE_WORD), "POS_BACKSPACE_MODE_WORD", "word" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("PosBackspaceMode"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

