/*
 * Copyright (C) 2022 The Phosh Developers
 *
 * SPDX-License-Identifier: GPL-3-or-later
 */
#pragma once

#include <glib.h>

G_BEGIN_DECLS

void pos_init (void);
void pos_uninit (void);

G_END_DECLS

