
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "phosh-osk-enums.h" */

GType phosh_osk_completion_mode_flags_get_type (void);
#define PHOSH_TYPE_OSK_COMPLETION_MODE_FLAGS (phosh_osk_completion_mode_flags_get_type())

GType phosh_osk_features_get_type (void);
#define PHOSH_TYPE_OSK_FEATURES (phosh_osk_features_get_type())

GType phosh_osk_completion_source_flags_get_type (void);
#define PHOSH_TYPE_OSK_COMPLETION_SOURCE_FLAGS (phosh_osk_completion_source_flags_get_type())

GType phosh_osk_scaling_flags_get_type (void);
#define PHOSH_TYPE_OSK_SCALING_FLAGS (phosh_osk_scaling_flags_get_type())

/* enumerations from "pos-enums.h" */

GType pos_osk_widget_layer_get_type (void);
#define POS_TYPE_OSK_WIDGET_LAYER (pos_osk_widget_layer_get_type())

GType pos_osk_widget_mode_get_type (void);
#define POS_TYPE_OSK_WIDGET_MODE (pos_osk_widget_mode_get_type())

GType pos_osk_key_use_get_type (void);
#define POS_TYPE_OSK_KEY_USE (pos_osk_key_use_get_type())

GType pos_input_method_text_change_cause_get_type (void);
#define POS_TYPE_INPUT_METHOD_TEXT_CHANGE_CAUSE (pos_input_method_text_change_cause_get_type())

GType pos_input_method_purpose_get_type (void);
#define POS_TYPE_INPUT_METHOD_PURPOSE (pos_input_method_purpose_get_type())

GType pos_input_method_hint_get_type (void);
#define POS_TYPE_INPUT_METHOD_HINT (pos_input_method_hint_get_type())

GType pos_backspace_mode_get_type (void);
#define POS_TYPE_BACKSPACE_MODE (pos_backspace_mode_get_type())

G_END_DECLS

/* Generated data ends here */

