#ifndef __RESOURCE_pos_H__
#define __RESOURCE_pos_H__

#include <gio/gio.h>

G_GNUC_INTERNAL GResource *pos_get_resource (void);

G_GNUC_INTERNAL void pos_register_resource (void);
G_GNUC_INTERNAL void pos_unregister_resource (void);

#endif
